/**********************************************************************
 *$Component                                                          *
 *    LOGICAL lt_global_keyword_check ()                              *
 *$Abstract                                                           *
 *    Checks for pointers in a PDS label                              *
 *$Keywords                                                           *
 *    LABEL_TOOL                                                      *
 *    KEYWORD                                                         *
 *    NON_UI_COMMON                                                   *
 *$Inputs                                                             *
 *    label_ptr:                                                      *
 *        The label_ptr variable is a pointer to the root of an       *
 *        ODL tree.  Usually it points to the actual "ROOT" object,   *
 *        but it may point to any other object in the tree.  If it    *
 *        does point to something other than the "ROOT" object, then  *
 *        this object will be treated as the root of a sub-tree, and  *
 *        processing will not be allowed to move above the level of   *
 *        this object.                                                *
 *$Outputs                                                            *
 *    None                                                            *
 *$Returns                                                            *
 *    success_flag:                                                   *
 *        The success_flag variable is a general purpose variable     *
 *        which indicates success (a value of TRUE) or failure        *
 *        (a value of FALSE).                                         *
 *$Detailed_Description                                               *
 *    The lt_global_keyword_check routine verifies that files         *
 *                 files specified by pointers actually exist.        *
 *                                                                    *
 *$External_References                                                *
 *    None                                                            *
 *$Author_and_Institution                                             *
 *    Dale Schultz/JPL                                                *
 *$Version_and_Date                                                   *
 *    1.0   January 9, 1998                                           *
 *$Change_History                                                     *
 *    MDD   01-09-98   Original Code.                                 *
 **********************************************************************/

LOGICAL lt_global_keyword_check (label_ptr)
AGGREGATE label_ptr;
{
   AGGREGATE temp_object_ptr = NULL;
   int status;
   int i;
   LOGICAL success = FALSE;

/** BEGIN **/

   /*------------------------------------------------------------------*/
   /** IF a keyword name and a label were provided THEN               **/
   /*------------------------------------------------------------------*/

   status = PDS_ERROR;
      /*---------------------------------------------------------------*/
      /** LOOP through all the objects in the label subtree           **/
      /*---------------------------------------------------------------*/

      temp_object_ptr = label_ptr;
      while (temp_object_ptr != NULL)
      {
         /*------------------------------------------------------------*/
         /** CASE operation of                                        **/
         /*------------------------------------------------------------*/
	 printf("%s\n", temp_object_ptr -> name);

         /*------------------------------------------------------------*/
         /** ENDCASE                                                  **/
         /*------------------------------------------------------------*/

	 temp_object_ptr = NextSubAggregate (label_ptr, temp_object_ptr);
      }
      /*---------------------------------------------------------------*/
      /** ENDLOOP                                                     **/
      /*---------------------------------------------------------------*/

   return (success);

/** END **/
}
/**********************************************************************
 *$Component                                                          *
 *    AGGREGATE lt_search_expanded_label (label_fname)                  *
 *$Abstract                                                           *
 *    Read STRUCTURE pointers and build an expanded tree.             *
 *$Keywords                                                           *
 *    LABEL_TOOL                                                      *
 *    STRUCTURE                                                       *
 *    NON_UI_COMMON                                                   *
 *$Inputs                                                             *
 *    label_fname:                                                    *
 *        The label_fname variable is a pointer to a character string *
 *        containing a file name.                                     *
 *$Outputs                                                            *
 *    None                                                            *
 *$Returns                                                            *
 *    None                                                            *
 *$Detailed_Description                                               *
 *    This routine takes a file containing a PDS label, builds a tree *
 *    from the ODL statements in the file, searches the tree for all  *
 *    STRUCTURE keywords, builds trees from the label files pointed   *
 *    to by the STRUCTURE keywords, merges the new trees into the     *
 *    parent tree where it found the STRUCTURE keyword, continues     *
 *    searching the tree, including the newly merged part, for more   *
 *    STRUCTURE keywords to expand.  The file names extracted from    *
 *    the STRUCTURE keywords are appended onto a STRING_LIST which    *
 *    is attached to the root node of the main tree by using the      *
 *    "appl2" field.  The "appl2" field in each child of the root,    *
 *    including the expanded and merged parts, is used to point to    *
 *    the text part of the STRING_LIST element which has the file name*
 *    where the objects were pulled in from.                          *
 *$External_References                                                *
 *    Item                       Shared-Data               Access     *
 * ------------------------------------------------------------------ *
 *  pds_verbose                  pdsglob.h                 write      *
 *  pds_watch_ends               pdsglob.h                 write      *
 *  pds_last_message             pdsglob.h                 write      *
 *$Author_and_Institution                                             *
 *    David P. Bernath / JPL                                          *
 *$Version_and_Date                                                   *
 *    2.0   March 3, 1992                                             *
 *$Change_History                                                     *
 *    DPB   12-04-91   Original Code.                                 *
 *    MDD   03-02-92   Corrected bugs resulting when structure files  *
 *                     could not be found or when more than one       *
 *                     structure pointer existed at the ROOT level.   *
 *                     Added code to turn off end statement checking  *
 *                     and to add file name indicators to the error   *
 *                     list.                                          *
 **********************************************************************/

AGGREGATE lt_search_expanded_label (label_fname)

char *label_fname;

{
    AGGREGATE label_ptr = {NULL};
    AGGREGATE sub_label_ptr = {NULL};
    AGGREGATE object_ptr = {NULL};
    AGGREGATE parent_ptr = {NULL};
    char err_msg [PDS_MAXLINE];
    char *structure_fname = {NULL};
    int i;
    LOGICAL found = {FALSE};
    LOGICAL save_watch;
    ERROR_LIST *message_ptr = {NULL};
    LOGICAL save_verbose;
    char savefile [PDS_MAXLINE];
    char *newfile = {NULL};
    char *label_dir = {NULL};
    char *struct_dir = {NULL};
    FILE *f_ptr = NULL;

/** BEGIN **/

    /*-----------------------------------------------------------------------*/
    /** Save the previous "end watch" and verbose flags                     **/
    /** Add a message header for this file                                  **/
    /** Save the directory name of the label file                           **/
    /** Construct the main tree from the file name passed in.               **/
    /** Remove the message header if the label had no errors.               **/
    /** Add a message trailer if the label had errors.                      **/
    /** Attach the file name to all the objects in the main tree.           **/
    /*-----------------------------------------------------------------------*/

    save_watch = pds_watch_ends;
    save_verbose = pds_verbose;
    label_dir = sys_get_path (label_fname);
    sprintf (err_msg, "***** Messages from LABEL file: %s *****", label_fname);
    err_append_message (CONTINUE, err_msg);
    message_ptr = pds_last_message;
    label_ptr = lab_read_label_or_template (label_fname);
    if (message_ptr == pds_last_message)
       err_deallocate_list (message_ptr);
    else
    {
       sprintf (err_msg, 
                "***** End of messages for LABEL file: %s *****", 
                label_fname);
       err_append_message (CONTINUE, err_msg);
    }
    ltx_attach_file_name (label_ptr, label_fname);

    /*-----------------------------------------------------------------------*/
    /** LOOP through the objects in the main tree, including any that have  **/
    /**         been merged in from previous STRUCTURE keywords.            **/
    /*-----------------------------------------------------------------------*/

    for (object_ptr = label_ptr; 
             object_ptr != NULL; object_ptr = NextAggregate(object_ptr))
    {
        /*-------------------------------------------------------------------*/
        /** Turn off the sfdu warnings                                      **/
        /** Turn off the "watch end statements" flag.                       **/
        /** LOOP through all the STRUCTURE keywords in the current object.  **/
        /*-------------------------------------------------------------------*/

        pds_verbose = FALSE;
        pds_watch_ends = FALSE;
        for (i=1; (structure_fname = lu_keyword_value (object_ptr,
                                         "STRUCTURE",i,FALSE)) != NULL; ++i)
        {    
            if ((f_ptr = fopen (structure_fname, "r")) == NULL)
            {
                strcpy (savefile, structure_fname);
                util_lower_case (structure_fname);
                if ((f_ptr = fopen (structure_fname, "r")) == NULL)
                {
                   if ((struct_dir = sys_get_path (structure_fname)) == NULL)
                   {
                       newfile = util_create_file_spec (label_dir, savefile);
                       Replace_String(structure_fname, newfile);
                       Lemme_Go(newfile);
                       if ((f_ptr = fopen (structure_fname, "r")) == NULL)
                          util_lower_case (structure_fname);
				   }
                    Lemme_Go(struct_dir);
            }
	    }
             if (f_ptr) fclose (f_ptr);

            /*---------------------------------------------------------------*/
            /** Determine if the file name attached to the current STRUCTURE**/
            /**     keyword is being included recursively (e.g., one of     **/
            /**     its parent objects came from the same file).            **/
            /*---------------------------------------------------------------*/

            for (parent_ptr=object_ptr->parent, found=FALSE;
                    ((parent_ptr != NULL) && (parent_ptr != label_ptr) && (! found)); 
                        parent_ptr = parent_ptr -> parent)
            {
                found = (strcmp (structure_fname, 
                                     (char *) (parent_ptr->appl2)) == 0);
            }

            /*---------------------------------------------------------------*/
            /** IF this is a recursive STRUCTURE reference THEN             **/
            /**     Append a message onto the list of messages.             **/
            /** ELSE                                                        **/
            /*---------------------------------------------------------------*/

            if (found)
            {
                sprintf (err_msg, 
                  "Recursive STRUCTURE reference found, involving the file: %s",
                   structure_fname);
		err_append_message (WARNING, err_msg);
            }
            else
            {
                /*-----------------------------------------------------------*/
                /** Create a message header for this structure file         **/
                /** Build a temp tree from the STRUCTURE file name.         **/
                /** Remove the message header if there were no errors in    **/
                /**    the label.                                           **/
                /** Add a message trailer if the label had errors.          **/
                /*-----------------------------------------------------------*/

                sprintf (err_msg, 
                         "***** ODL syntax messages for STRUCTURE file: %s *****", 
                         structure_fname);
		err_append_message (CONTINUE, err_msg);
                message_ptr = pds_last_message;
                sub_label_ptr = lab_read_label_or_template (structure_fname);
                if (message_ptr == pds_last_message)
		   err_deallocate_list (message_ptr);
                else
                {
                    sprintf (err_msg, 
                         "***** End of messages for STRUCTURE file: %s *****", 
                         structure_fname);
		    err_append_message (CONTINUE, err_msg);
		}

                /*-----------------------------------------------------------*/
                /** IF a structure file was found THEN                      **/
                /**    Attach the file name to all objects in the temp tree **/
                /**    Append the temp tree's file name list onto the main  **/
                /**        tree's list.                                     **/
                /**    Merge the temp tree into the main tree where the     **/
                /**        STRUCTURE keyword was found.                     **/
                /**    Remove the temp tree.                                **/
                /*-----------------------------------------------------------*/

                if (sub_label_ptr != NULL)
                {
                   ltx_attach_file_name (sub_label_ptr, structure_fname);
                   label_ptr->appl2 = (long) util_append_string_list (
                                           (STRING_LIST *) (label_ptr->appl2), 
					   (char *) (sub_label_ptr->appl2),
					    LIST_TYPE);
                   ltx_merge_objects (sub_label_ptr, object_ptr);
                   lab_remove_label_or_template (sub_label_ptr);
		}
                /*-----------------------------------------------------------*/
                /** ELSE                                                    **/
                /**  append "NULL_FILE" onto the list of file names         **/
                /**  notify the user of the error                           **/
                /*-----------------------------------------------------------*/

                else
                {
                   label_ptr->appl2 = (long) util_append_string_list (
                                           (STRING_LIST *) (label_ptr->appl2), 
                                           "NULL_FILE", STRING_TYPE);
                   sprintf (err_msg, 
                     "Missing or bad STRUCTURE file: %s", structure_fname);
		   err_append_message (ERROR, err_msg);

                /*-----------------------------------------------------------*/
                /** ENDIF a structure file was found...                     **/
                /*-----------------------------------------------------------*/

                }

            /*---------------------------------------------------------------*/
            /** ENDIF                                                       **/
            /*---------------------------------------------------------------*/

            }  /*  End:  "if (found) ... else ..."  */

            Lemme_Go(structure_fname);

        /*-------------------------------------------------------------------*/
        /** ENDLOOP                                                         **/
        /*-------------------------------------------------------------------*/

        }  /*  End:  "for (i=1; (structure_fname = ..."  */

    /*-----------------------------------------------------------------------*/
    /** ENDLOOP                                                             **/
    /*-----------------------------------------------------------------------*/

    }  /*  End:  "for (object_ptr = label_ptr; ..."  */

    /*-----------------------------------------------------------------------*/
    /** RETURN a pointer to the root of the expanded tree.                  **/
    /*-----------------------------------------------------------------------*/
    
    Lemme_Go(label_dir);
    pds_watch_ends = save_watch;
    pds_verbose = save_verbose;
    return (label_ptr);

/** END **/

}  /*  "lt_read_expanded_label"  */
